/*
 * (C) Copyright 2003
 * Tait Electronics Limited, Christchurch, New Zealand
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <config.h>
#include <mpc8xx.h>
#include <i2c.h>
#include <asm-ppc/processor.h>

/* ------------------------------------------------------------------------- */
#define _not_used_ 0xffffffff


/* the bootcode for the dsp from RJ */
const ushort dspcode [] = 
{
0x761c,0x0088,0xfb01,0x2610, /*0x2610,0x2210 for 12.8/40MHz clock in */
0x9a76,0x3804,0x88a8,0x0199,
0xec31,0x9e00,0x00ef,0xc823,
0x46b0,0x3c18,0xc823,0x1283,
0x8004,0x18f8,0x2298,0xc823,
0x0419,0xf9ec,0x319e,0x0100,
0x00fb,0x2311,0x1122,0x98c8,
0x2304,0x19f9,0xec31,0x9e02,
0x0000,0xfb23,0x2222,0x2298,
0xc823,0x0419,0xf940,0x1862,
0x5846,0xc64a,0x0246,0xc746,
0xd246,0xd346,0xd246,0xd346,
0xd246,0xd34a,0x6820,0x0000
};

#define CODE_LENGTH	(sizeof(dspcode) / sizeof(dspcode[0]))


/* the bootcode is not located on the reset vector so redirect */
const ushort dspvector [] = 
{
0x6A00,0x00C0
};
#define VECTOR_LENGTH	(sizeof(dspvector) / sizeof(dspvector[0]))


#define DSP_TIME_OUT		55
#define DSP_TIME_UNDER		40
#define DSP_TIMER_SAMPLES	10

int test_last_reset_magic(void);
static long int dram_size (long int mamr_value, long int *base, long int maxsize);
static long int load_dsp_code(const ushort *address);




/*
 * Check Board Identity:
 */

int checkboard (void)
{
	volatile immap_t     *immap = (immap_t *)CFG_IMMR;
	volatile memctl8xx_t *memctl = &immap->im_memctl;
	ulong immr, data;
	int partnum, masknum;
	long int dsp_mask_no;

	
	/*
	 * Map the DSP HPI into our memory space via Chip Select 2
	 */
	memctl->memc_or2 = CFG_OR2_DSPHPI & ~OR_SETA;	/* use internal TA to start with */
	memctl->memc_br2 = CFG_BR2_DSPHPI;
	 
	
	immr = mfspr(638);
	partnum = (immr >> 8) & 0xff;
	masknum = immr & 0xff;
	
	printf("CPU Part = %04X: Microcode Mask = %04X: ",partnum,masknum);
	switch (immr & 0xffff) {
	case 0x800:
	    printf("Rev Number = 0.3\n");
	    break;
	case 0x801:
	    printf("Rev Number = A.0\n");
	    break;
	default:
	    printf("Unknown CPU rev number\n");
	}

	puts ("Board: APCO System InterFace ");
#if CONFIG_SPIN1
	puts ("spin 1 build\n");
#endif
#if CONFIG_SPIN2
	puts ("spin 2 build\n");
#endif	
#if CONFIG_SPIN3
	puts ("spin 3 build\n");
#endif	

        /*
	 * Release reset on the DSP chip which hangs off PA13
	 * by making it an output and driving it high
	 */
	immap->im_ioport.iop_papar &= ~PA_DSP_RESET;	/* assignment = general i/o */
	immap->im_ioport.iop_padir |=  PA_DSP_RESET;	/* direction = out */
	immap->im_ioport.iop_padat |=  PA_DSP_RESET;	/* output = high */

	/*
	 * Set as active the ASIF Not Ready line
	 */

	immap->im_ioport.iop_papar &= ~PA_ASIF_NRDY;	/* assignment = general i/o */
	immap->im_ioport.iop_padir |=  PA_ASIF_NRDY;	/* direction = out */
	immap->im_ioport.iop_padat &= ~PA_ASIF_NRDY;	/* output = low */


	/*
	 * Load the test dsp code and get the mask version of the dsp
	 */
	dsp_mask_no = load_dsp_code(dspcode);

	/*
	 * The returned value will either be negative to indicate there is a problem with the 
	 * dsp or the mask version number
	 */
	switch (dsp_mask_no) {
	case -1:
	        puts ("DSP memory error\n");
	        break;
	case -2:
	        puts ("DSP load/run error\n");
	        break;
	case -3:
		puts ("DSP HINT line toggling too slow\n");
		break;
	case -4:
		puts ("DSP HINT line toggling too fast\n");
		break;
	default:
		printf("DSP loaded OK. Mask version %04X\n", dsp_mask_no);
	}

#if CONFIG_SPIN3
	/* switch to external Transfer Acknowledge signal */
	if ((data = getenv ("ext_ta")) != NULL)
		memctl->memc_or2 = CFG_OR2_DSPHPI | OR_SETA;
#endif
	/*
	 * Always initialise the codec, even if we never use it in the
	 * bootloader - handy for factory tests!!
	 */
	i2c_init (CFG_I2C_SPEED, CFG_I2C_SLAVE);

	data = 0x93000000;
	if (i2c_write(CFG_I2C_CODEC_ADDR, 0, 1, (char *)&data, 1) != 0) {
		puts ("CODEC write error\n");
	}
	data = 0x03000000;
	if (i2c_write(CFG_I2C_CODEC_ADDR, 1, 1, (char *)&data, 1) != 0) {
		puts ("CODEC write error\n");
	}


        /* 
	 * Set one of the TOUT lines from the CPM buzzing at 25MHz to be
	 * the clock for the FEC PHY
	 */
#if CONFIG_SPIN1
	immap->im_cpmtimer.cpmt_tgcr &= ~TGCR_RST4 | TGCR_STP4;
	immap->im_cpmtimer.cpmt_tmr4 = ((0 << TMR_PS_SHIFT) & TMR_PS_MSK)
								 | TMR_FRR | TMR_ICLK_IN_GEN;
	immap->im_cpmtimer.cpmt_tcn4 = 0;
	immap->im_cpmtimer.cpmt_ter4 = 0xffff;
	immap->im_cpmtimer.cpmt_tgcr |= TGCR_RST4;
	immap->im_cpmtimer.cpmt_trr4 = (ushort)4;
	immap->im_cpmtimer.cpmt_tgcr &= ~TGCR_STP4;

#define PA_TOUT4	0x8000
	immap->im_ioport.iop_papar |=  PA_TOUT4;	/* assignment = peripheral */
        immap->im_ioport.iop_padir |=  PA_TOUT4;	/* direction = out */
#endif
/* spin 2 boards use tout1 */
#if CONFIG_SPIN2 || CONFIG_SPIN3
	immap->im_cpmtimer.cpmt_tgcr &= ~TGCR_RST1 | TGCR_STP1;
	immap->im_cpmtimer.cpmt_tmr1 = ((0 << TMR_PS_SHIFT) & TMR_PS_MSK)
								 | TMR_FRR | TMR_ICLK_IN_GEN;
	immap->im_cpmtimer.cpmt_tcn1 = 0;
	immap->im_cpmtimer.cpmt_ter1 = 0xffff;
	immap->im_cpmtimer.cpmt_tgcr |= TGCR_RST1;
	immap->im_cpmtimer.cpmt_trr1 = (ushort)4;	/* 125MHz / 5 = 25MHz */
	immap->im_cpmtimer.cpmt_tgcr &= ~TGCR_STP1;

#define PA_TOUT1	0x0200
	immap->im_ioport.iop_papar |=  PA_TOUT1;	/* assignment = peripheral */
        immap->im_ioport.iop_padir |=  PA_TOUT1;	/* direction = out */
#endif

	*(volatile ulong *)(immap->im_cpm.cp_dpmem + CFG_CPM_LAST_RESET_ADDR) = test_last_reset_magic();

	return 0;

}



/* waits for an interrupt (polling) */
static int wait_for_dsp_hint(void)
{
	volatile immap_t     *immap = (immap_t *)CFG_IMMR;
	long timeout;
	uint dp0;
	/* we are running from Flash so need a static variable */
	DECLARE_GLOBAL_DATA_PTR;

	for (timeout = 0; timeout < DSP_TIME_OUT; timeout++) {
		dp0 = (immap->im_siu_conf.sc_sipend & 0x02000000);
		if (dp0 != gd->dp0_last) {
			gd->dp0_last = dp0;
			return timeout;
		}
		udelay(100);
	}
	return -1;
}



/* note delay after write to allow for DSP running at default clock */
void dsp_write_memory(int offset, ushort value)
{
        vu_short *load_address = (ushort *)CFG_DSPHPI_BASE;


	if (offset & 1) {
		load_address[(offset & ~1)/CFG_DSPHPI_INC + CFG_DSPHPI_OFFSET] = value;
		udelay(5);
	} else {
		load_address[offset/CFG_DSPHPI_INC] = value;
		udelay(5);
	}
}

/* Only likely to work when DSP at full clock speed!! */
ushort dsp_read_memory(int offset)
{
        vu_short *load_address = (ushort *)CFG_DSPHPI_BASE;
	vu_short value;

	if (offset & 1) {
		value = load_address[(offset & ~1)/CFG_DSPHPI_INC + CFG_DSPHPI_OFFSET];
	} else {
		value = load_address[offset/CFG_DSPHPI_INC];
	}
	return value;
}

/* -------------------------------------------------------------------------
 * load DSP with basic bootstrap code to get its clocks running at the  rate. 
 * Use the defines CFG_DSPHPI_LOAD, CFG_DSPHPI_OFFSET & CFG_DSPHPI_INC to suss
 * out the addressing on spin one PCBs, start dsp from CFG_DSP_CONTROL
 *  Memory loaded by this test software (by Word address):
 *  0x00060 - 0x0008F This application code
 *  0x000EF - 0x000EF Silicon Revision 0x6512==Revision 2.2
 *  0x000F0 - 0x000FF Sequence 0x0001, 0x02,0x04,0x08,0x10... 0x8000
 *  0x00100 - 0x27FFF Sequence containing ls 16bits of word address
 *  0x10000 has a special value 0x1111
 *  0x20000 has a special value 0x2222
 *  Output 3 low going pulses on external pin XF for ever (approx 6.5ns wide)
 *
 * return values as follows:
 * -1: can't find any dsp memory
 * -2: can't find correct dsp pattern
 * other: dsp mask version
 */

static long int load_dsp_code(const ushort *code_addr)
{

	uint i;
	int samples[DSP_TIMER_SAMPLES];
	ushort data;

	/*
	 * write dsp vector code to the correct place
	 */
	for (i = 0; i < VECTOR_LENGTH; i++) {
		dsp_write_memory(CFG_DSPHPI_VECTOR + i, dspvector[i]);
	}

	/*
	 * write dsp main code to the correct place
	 * do this after the vector in case we run over the vector deliberately
	 */
	for (i = 0; i < CODE_LENGTH; i++) {
		dsp_write_memory(CFG_DSPHPI_LOAD + i, code_addr[i]);
	}
		
	/*
	 * write to dsp control word to start the DSP
	 */
	dsp_write_memory(CFG_DSPHPI_CONTROL, CFG_DSPHPI_START_CODE);

        udelay(100000);				/* wait for DSP to do its stuff */

	/*
	 * check memory by ensuring we can read back the original code
	 * Reads should work reliably now we have the DSP at full speed
	 */
	for (i = 0; i < CODE_LENGTH; i++) {
		if ((data = dsp_read_memory(CFG_DSPHPI_LOAD + i)) != dspcode[i]) { 
			printf("Expected %04X, got %04X at offset %d\n", dspcode[i], data, i);
			return -1;
		}
	}

	/*
	 * Check a few addresses to see that the DSP has written its pattern
	 * as we expect and then get the mask version number from a known location
	 */
	if ((dsp_read_memory(CFG_DSPHPI_ADDR1) != CFG_DSPHPI_CODE1) ||
	          (dsp_read_memory(CFG_DSPHPI_ADDR2) != CFG_DSPHPI_CODE2)) {
		return -2;
	}

	/*
	 * Check the interval of the toggling that should be happening on the HINT
	 * line from the DSP. This appears on DP0_IRQ3 and should be about 4/4.5mS each way
	 */
	/* read several intervals successively, ignore 1st if too short */	 

	for (i = 0; i < DSP_TIMER_SAMPLES; i++) {
		samples[i] = wait_for_dsp_hint();	/* save samples and check/print later */
	}
    
	for (i = 0; i < DSP_TIMER_SAMPLES; i++) {
		debug ("HINT period %d = %d00uS\n", i, samples[i]);
		if ( samples[i] < 0) {
			return -3;			/* interval too long */
		}
		if (( samples[i] < DSP_TIME_UNDER) && (i > 0)) {
			return -4;			/* interval too short */
		}
	}

	return (long) dsp_read_memory(CFG_DSPHPI_VERSION);

}






/* ------------------------------------------------------------------------- */
/* sdram table stolen from the fads manual                                   */
/* for chip MB811171622A-100                                                 */

/* this table is for 32-50MHz operation */


const uint sdram_table[] =
{
	/* single read. (offset 0 in upm RAM) */
	0x1f07fc04, 0xeeaefc04, 0x11adfc04, 0xefbbbc00,
	0x1ff77c47,

	/* MRS initialization (offset 5) */

	0x1ff77c34, 0xefeabc34, 0x1fb57c35,

	/* burst read. (offset 8 in upm RAM) */
	0x1f07fc04, 0xeeaefc04, 0x10adfc04, 0xf0affc00,
	0xf0affc00, 0xf1affc00, 0xefbbbc00, 0x1ff77c47,
	_not_used_, _not_used_, _not_used_, _not_used_,
	_not_used_, _not_used_, _not_used_, _not_used_,

	/* single write. (offset 18 in upm RAM) */
	0x1f27fc04, 0xeeaebc00, 0x01b93c04, 0x1ff77c47,
	_not_used_, _not_used_, _not_used_, _not_used_,

	/* burst write. (offset 20 in upm RAM) */
	0x1f07fc04, 0xeeaebc00, 0x10ad7c00, 0xf0affc00,
	0xf0affc00, 0xe1bbbc04, 0x1ff77c47, _not_used_,
	_not_used_, _not_used_, _not_used_, _not_used_,
	_not_used_, _not_used_, _not_used_, _not_used_,

	/* refresh. (offset 30 in upm RAM) */
	0x1ff5fc84, 0xfffffc04, 0xfffffc04, 0xfffffc04,
	0xfffffc84, 0xfffffc07, _not_used_, _not_used_,
	_not_used_, _not_used_, _not_used_, _not_used_,

	/* exception. (offset 3c in upm RAM) */
	0x7ffffc07, _not_used_, _not_used_, _not_used_ };

/* ------------------------------------------------------------------------- */

long int initdram (int board_type)
{
	volatile immap_t     *immap = (immap_t *)CFG_IMMR;
	volatile memctl8xx_t *memctl = &immap->im_memctl;
	ulong size_cs1, size8, size9, size10;


        /* 
	 * Start by loading the UPM with the patterns required 
	 */
	upmconfig(UPMA, (uint *)sdram_table,sizeof(sdram_table)/sizeof(uint));


	/*
	 * Preliminary prescaler for refresh (depends on number of
	 * banks): This value is selected for 8192 cycles every 64 ms
	 * i.e every 15.6uS with one SDRAM bank. 
	 * Rethink this if major changes to memory size!!
	 */
	memctl->memc_mptpr = CFG_MPTPR_1BK_8K;

	memctl->memc_mamr = CFG_MAMR_8COL & (~(MAMR_PTAE));	/* no refresh yet */

	/*
	 * The following value is used as an address (i.e. opcode) for
	 * the LOAD MODE REGISTER COMMAND during SDRAM initialisation. If
	 * the port size is 32bit the SDRAM does NOT "see" the lower
	 * address lines, i.e. mar=0x00000022 -> opcode=0x00000088 for
	 * 16 bit wide SDRAMs:
	 * ->    0 00 010 0 010
	 *       |  |   | |   +- Burst Length = 4
	 *       |  |   | +----- Burst Type   = Sequential
	 *       |  |   +------- CAS Latency  = 2
	 *       |  +----------- Operating Mode = Standard
	 *       +-------------- Write Burst Mode = Programmed Burst Length
	 */
	memctl->memc_mar = 0x00000088;

	/* 
	 * Now run the precharge/nop/mrs commands.
	 */

	memctl->memc_mcr = 0x80002111;   /* run pattern 0x11 (0xffffffff) once */
	udelay(200);

	/* Perform SDRAM initializsation sequence from location 5
	 *
	 * 10xxx xxx0x xxxxx x001x 0001 xx00 0101
	 *  |       |           |     |  |      +- memory array address
	 *  |       |           |     |  +-------- unspec = 0
	 *  |       |           |     +----------- loop count
	 *  |       |           +----------------- memory bank = CS1
	 *  |       +----------------------------- UPM = A
	 *  +------------------------------------- command = run
	 */
	memctl->memc_mcr = 0x80002105;	/* SDRAM bank 0 */
	udelay (200);

        /*
	 * Now run 8 refresh sequences (from location 30) to complete 
	 * the initialisation
	 */
	memctl->memc_mcr = 0x80002830;	/* SDRAM bank 0 - execute 8x */
	udelay (200);


	memctl->memc_mamr |= MAMR_PTAE;	/* enable refresh */
	udelay (200);

	/*
	 * Map controller bank 1 to the SDRAM bank at
	 * preliminary address - these have to be modified after the
	 * SDRAM size has been determined.
	 */
	memctl->memc_or1 = CFG_OR1_PRELIM;
	memctl->memc_br1 = CFG_BR1_PRELIM;

	/*
	 * Check CS1 Memory Size for re-configuration
	 *
	 */
	/* start with 8 column mode */
	size8 = dram_size (CFG_MAMR_8COL, (ulong *) SDRAM_BASE1_PRELIM, SDRAM_MAX_SIZE);
	debug("8 Column mode found 0x%0X bytes\n", size8);
	/* next try 9 column mode */
	size9 = dram_size (CFG_MAMR_9COL, (ulong *) SDRAM_BASE1_PRELIM, SDRAM_MAX_SIZE);
	debug("9 Column mode found 0x%0X bytes\n", size9);
	/* finally try 10 column mode */
	size10 = dram_size (CFG_MAMR_10COL, (ulong *) SDRAM_BASE1_PRELIM, SDRAM_MAX_SIZE);
	debug("10 Column mode found 0x%0X bytes\n", size10);

	/* 
	 * Choose the mode that gives the highest result
	 * This should be the correct mode for the devices fitted!!
	 */
	if (size9 > size8) {
		size_cs1 = size9;
		memctl->memc_mamr = CFG_MAMR_9COL;
	} else if (size10 > size9) {
		size_cs1 = size10;
		memctl->memc_mamr = CFG_MAMR_10COL;
	} else {
		size_cs1 = size8;
		memctl->memc_mamr = CFG_MAMR_8COL;
	}


	/*
	 * Adjust mapping now we know how big the RAM chip is
	 */
	memctl->memc_or1 = ((-size_cs1) & OR_AM_MSK) | CFG_OR_TIMING_SDRAM;
	memctl->memc_br1 = ((CFG_SDRAM_BASE & BR_BA_MSK) | BR_MS_UPMA | BR_V);

	/* 
	 * return size to caller
	 */
	return size_cs1;
}

/* ------------------------------------------------------------------------- */

/*
 * Check memory range for valid RAM. A simple memory test determines
 * the actually available RAM size between addresses `base' and
 * `base + maxsize'. Some (not all) hardware errors are detected:
 * - short between address lines
 * - short between data lines
 * For this to work, the address mask for this chip select must be 
 * bigger than 'maxsize'
 */

static long int dram_size (long int mamr_value, long int *base,
						   long int maxsize)
{
	volatile immap_t *immap = (immap_t *) CFG_IMMR;
	volatile memctl8xx_t *memctl = &immap->im_memctl;
	volatile long int *addr;
	ulong cnt, val;
	ulong save[32];				/* to make test non-destructive */
	unsigned char i = 0;

	memctl->memc_mamr = mamr_value;

	for (cnt = maxsize / sizeof (long); cnt > 0; cnt >>= 1) {
		addr = base + cnt;		/* pointer arith! */

		save[i++] = *addr;
		*addr = ~cnt;
	}

	/* write 0 to base address */
	addr = base;
	save[i] = *addr;
	*addr = 0;

	/* check at base address */
	if ((val = *addr) != 0) {
		*addr = save[i];
		return (0);
	}

	for (cnt = 1; cnt <= maxsize / sizeof (long); cnt <<= 1) {
		addr = base + cnt;		/* pointer arith! */

		val = *addr;
		*addr = save[--i];

		if (val != (~cnt)) {
			return (cnt * sizeof (long));
		}
	}
	return (maxsize);
}

/* ------------------------------------------------------------------------- */

/* get all digital input values to the ppc and scrunch them into a single 16 bit */
short get_digio_value(void)
{
	int	irq, onepps, portc, pip;
	volatile pcmconf8xx_t *pcmp = (pcmconf8xx_t *)(&(((immap_t *)CFG_IMMR)->im_pcmcia));
	volatile immap_t *immr = (immap_t *) CFG_IMMR;


	/* look for IRQ 1 (1pps) and IRQ 6 (E line) */
	irq  = ((~immr->im_siu_conf.sc_sipend & 0x20000000) >> 29);
	irq |= ((~immr->im_siu_conf.sc_sipend & 0x00080000) >> 18);
	/* 1pps and E line also enter on port pins */
	onepps = ((immr->im_ioport.iop_padat & 0x1000) >> 12);
        portc = immr->im_ioport.iop_pcdat & 0x0f;
	pip  = ((pcmp->pcmc_pipr & 0xff000000) >> 24);
	/*        1111 11	   */ 
	/*        5432 1098 7654 3210 */
	/* return xiio cccc pppp pppp */
	return ((irq << 13) | (onepps << 12) | (portc << 8) | (pip));

}



/* ------------------------------------------------------------------------- */
/* This doesn't actually show anything - it stores the values in memory */
void show_boot_progress(ulong status)
{

	volatile void *save_addr =
		((immap_t *) CFG_IMMR)->im_cpm.cp_dpmem + CFG_CPM_PROGRESS_WORD_ADDR;

	*(volatile ulong *) save_addr   = status;


}


/* ------------------------------------------------------------------------- */
/* Return true if last reset was just a reset and not a power-up */
int test_last_reset_magic(void)
{
#define BOOT_MAGIC1	0xA51FF15A
#define BOOT_MAGIC2	0xDEADBEEF
	ulong value1;
	ulong value2;

	volatile void *save_addr =
		((immap_t *) CFG_IMMR)->im_cpm.cp_dpmem + CFG_CPM_MAGIC_WORD_ADDR;

	value1 = *(volatile ulong *) (save_addr + 0);
	value2 = *(volatile ulong *) (save_addr + 4);

	*(volatile ulong *) (save_addr + 0) = BOOT_MAGIC1;
	*(volatile ulong *) (save_addr + 4) = BOOT_MAGIC2;
	
	if ((value1 == BOOT_MAGIC1) && (value2 == BOOT_MAGIC2)) {
		return 1;
	} else {
		return 0;
	}
		
}


/* ------------------------------------------------------------------------- */
/* not yet implemented */
int testdram (void)
{
    /* TODO: XXX XXX XXX */
    printf ("test not yet implemented\n");

    return (0);
}



/* ------------------------------------------------------------------------- */
/* stub function to void hacking board.c */
void board_poweroff (void)
{

    while (1);
}

/* ------------------------------------------------------------------------- */
/* stub function to void hacking post.c */
int post_hotkeys_pressed (gd_t *gd)
{

    return 0;			/* no hotkeys supported */
}

/* ------------------------------------------------------------------------- */
