/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 * Parts Copyright 2003 Tait Electronics Limited, Christchurch, New Zealand
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <mpc8xx.h>

flash_info_t	flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips	*/

#if defined(CFG_ENV_IS_IN_FLASH)
# ifndef  CFG_ENV_ADDR
#  define CFG_ENV_ADDR	(CFG_FLASH_BASE + CFG_ENV_OFFSET)
# endif
# ifndef  CFG_ENV_SIZE
#  define CFG_ENV_SIZE	CFG_ENV_SECT_SIZE
# endif
# ifndef  CFG_ENV_SECT_SIZE
#  define CFG_ENV_SECT_SIZE  CFG_ENV_SIZE
# endif
#endif

#define FLASH_DATA_MASK 0xffff
#define FLASH_MANU_MASK 0x00ff


/*-----------------------------------------------------------------------
 * Functions
 */
static ulong flash_get_size (vu_long *addr, flash_info_t *info);
static int write_word (flash_info_t *info, ulong dest, ulong data);
static void flash_get_offsets (ulong base, flash_info_t *info);
static int write_halfword (flash_info_t *info, ulong dest, ushort data);
static int _flash_protect(flash_info_t *info, ulong sector);
static int _flash_unprotect(flash_info_t *info, ulong sector);

/*-----------------------------------------------------------------------
 */

unsigned long flash_init (void)
{
//	volatile immap_t	*immap  = (immap_t *)CFG_IMMR;
//	volatile memctl8xx_t	*memctl = &immap->im_memctl;
	unsigned long size;
	int i;


	for (i=0; i<CFG_MAX_FLASH_BANKS; ++i) {
		flash_info[i].flash_id = FLASH_UNKNOWN;
	}

	/* Static FLASH Bank configuration here */

	size = flash_get_size((vu_long *)CFG_FLASH_BASE, &flash_info[0]);

	if (flash_info[0].flash_id == FLASH_UNKNOWN) {
		printf ("## Unknown FLASH on Bank 0 - Size = 0x%08lx = %ld MB\n",
			size, size<<20);
	}

	/* Remap FLASH according to real size */
	/* DON'T!! keep larger size for multiple flash sizes */
//	memctl->memc_or0 = CFG_OR_TIMING_FLASH | (-size & 0xFFFF8000);
//	memctl->memc_br0 = (CFG_FLASH_BASE & BR_BA_MSK) | (memctl->memc_br0 & ~(BR_BA_MSK));



/* unlock all of flash, then lock only the bits we want locking */
	flash_protect(FLAG_PROTECT_CLEAR,
			CFG_FLASH_BASE,
			CFG_FLASH_BASE + size - 1,
			&flash_info[0]);

/* protect the sector above the environment for external progs storage */
#ifdef CFG_ENV_ADDR_REDUND
#define CFG_EXTERNAL_PROG (CFG_ENV_ADDR_REDUND + CFG_ENV_SECT_SIZE)
#else
#define CFG_EXTERNAL_PROG (CFG_ENV_ADDR + CFG_ENV_SECT_SIZE)
#endif

	flash_protect(FLAG_PROTECT_SET,
		      CFG_EXTERNAL_PROG,
		      CFG_EXTERNAL_PROG + CFG_ENV_SECT_SIZE - 1,
		      &flash_info[0]);

#if CFG_MONITOR_BASE >= CFG_FLASH_BASE
	/* monitor protection ON by default */
	flash_protect(FLAG_PROTECT_SET,
		      CFG_FLASH_BASE + size - CFG_MONITOR_OFFSET,
		      CFG_FLASH_BASE + size - CFG_MONITOR_OFFSET + CFG_MONITOR_LEN - 1,
		      &flash_info[0]);
#endif

#ifdef	CFG_ENV_IS_IN_FLASH
	/* ENV protection ON by default */
	flash_protect(FLAG_PROTECT_SET,
		      CFG_ENV_ADDR,
		      CFG_ENV_ADDR + CFG_ENV_SECT_SIZE - 1,
		      &flash_info[0]);
#endif

#ifdef CFG_ENV_ADDR_REDUND
	flash_protect(FLAG_PROTECT_SET,
		      CFG_ENV_ADDR_REDUND,
		      CFG_ENV_ADDR_REDUND + CFG_ENV_SECT_SIZE - 1,
		      &flash_info[0]);
#endif
	return (size);
}

/*-----------------------------------------------------------------------
 */
static void flash_get_offsets (ulong base, flash_info_t *info)
{
	int i;

	/* set up sector start address table */
	if (info->flash_id & FLASH_BTYPE)
	{	/* bottom boot sector types - these are the useful ones! */
		/* set sector offsets for bottom boot block type */
		switch (info->flash_id & FLASH_TYPEMASK) {
		case FLASH_28F320C3B:
		case FLASH_28F640C3B:
		case FLASH_STMW320CB:
		case FLASH_STMW640ECB:
			/* STMW320CB & STMW640ECB have 8 x 8k bottom boot sectors */
			for (i = 0; i < 8; i++)												/* +8k		*/
				info->start[i] = base + (i * (0x00002000));
			for (; i < info->sector_count; i++)									/* +64k		*/
				info->start[i] = base + (i * (0x00010000)) - (0x00070000);
			break;
		default:
			puts("Unknown Flash\n");
		}
	}
	else
	{	/* top boot sector types - not so useful */
		puts("Invalid top boot Flash\n");
	}
}

/*-----------------------------------------------------------------------
 */
void flash_print_info  (flash_info_t *info)
{
	int i;

	if (info->flash_id == FLASH_UNKNOWN) {
		printf ("missing or unknown FLASH type\n");
		return;
	}

	switch (info->flash_id & FLASH_VENDMASK) {
	case FLASH_MAN_AMD:	printf ("AMD ");		break;
	case FLASH_MAN_FUJ:	printf ("FUJITSU ");		break;
	case FLASH_MAN_STM:	printf ("ST ");			break;
	case FLASH_MAN_INTEL:	printf ("Intel ");		break;
	default:		printf ("Unknown Vendor ");	break;
	}

	switch (info->flash_id & FLASH_TYPEMASK) {
	case FLASH_AM320B:	printf ("AM29LV320B (32 Mbit, bottom boot sect)\n");
				break;
	case FLASH_STMW320CB:	printf ("M28W320CB (32 Mbit, bottom boot sect)\n");
				break;
	case FLASH_STMW640ECB:	printf ("M28W640ECB (64 Mbit, bottom boot sect)\n");
				break;
	case FLASH_28F320C3B:	printf ("28F320C3 (32 Mbit, bottom boot sect)\n");
				break;
	case FLASH_28F640C3B:	printf ("28F640C3 (64 Mbit, bottom boot sect)\n");
				break;
	default:		printf ("Unknown Chip Type\n");
				break;
	}

	printf ("  Size: %ld MB in %d Sectors\n",
		info->size >> 20, info->sector_count);

	printf ("  Sector Start Addresses:");
	for (i=0; i<info->sector_count; ++i) {
		if ((i % 5) == 0)
			printf ("\n   ");
		printf (" %08lX%s",
			info->start[i],
			info->protect[i] ? " (RO)" : "     "
		);
	}
	printf ("\n");
	return;
}

/*-----------------------------------------------------------------------
 */


/*-----------------------------------------------------------------------
 */

/*
 * The following code cannot be run from FLASH!
 */

static ulong flash_get_size (vu_long *address, flash_info_t *info)
{
	short i;
	ushort value;
	vu_short *addr = (vu_short *)address;
	ulong base = (ulong)address;

	/* Write "Intelligent Identifier" command: read Manufacturer ID */
	*addr = 0x0090;

	value = addr[0] & FLASH_MANU_MASK;
	switch (value) {
	case AMD_MANUFACT & FLASH_DATA_MASK:
		info->flash_id = FLASH_MAN_AMD;
		break;
	case FUJ_MANUFACT & FLASH_DATA_MASK:
		info->flash_id = FLASH_MAN_FUJ;
		break;
	case STM_MANUFACT & FLASH_DATA_MASK:
		info->flash_id = FLASH_MAN_STM;
		break;
	case MT_MANUFACT & FLASH_DATA_MASK:	/* MT or => Intel */
	case INTEL_ALT_MANU & FLASH_DATA_MASK:
		info->flash_id = FLASH_MAN_INTEL;
		break;
	default:
		info->flash_id = FLASH_UNKNOWN;
		info->sector_count = 0;
		info->size = 0;
		return (0);			/* no or unknown flash	*/
	}

	value = addr[1];	/* device ID		*/
	switch (value) {
	case AMD_ID_LV320B & FLASH_DATA_MASK:
		info->flash_id += FLASH_AM320B;
		info->sector_count = 71;
		info->size = 0x00400000;
		break;				/* => 4 or 8MB		*/

	case STM_ID_28W320CB & FLASH_DATA_MASK:
		info->flash_id += FLASH_STMW320CB;
		info->sector_count = 71;
		info->size = 0x00400000;
		break;				/* => 4 MB		*/

	case STM_ID_28W640ECB & FLASH_DATA_MASK:
		info->flash_id += FLASH_STMW640ECB;
		info->sector_count = 135;
		info->size = 0x00800000;
		break;				/* => 8 MB		*/

	case INTEL_ID_28F320C3B & FLASH_DATA_MASK:
		info->flash_id += FLASH_28F320C3B;
		info->sector_count = 71;
		info->size = 0x00400000;
		break;				/* => 2x4 MB		*/

	case INTEL_ID_28F640C3B & FLASH_DATA_MASK:
		info->flash_id += FLASH_28F640C3B;
		info->sector_count = 135;
		info->size = 0x00800000;
		break;				/* => 8 MB		*/


	default:
		info->flash_id = FLASH_UNKNOWN;
		return (0);			/* => no or unknown flash */

	}

	/* set up sector start address table */
	flash_get_offsets (base, info);

	/* check for protected sectors */
	for (i = 0; i < info->sector_count; i++) {
		/* read sector protection at sector address, (A7 .. A0) = 0x02 */
		/* D0 = 1 if protected */
		addr = (vu_short *)(info->start[i]);
		info->protect[i] = addr[2] & 1;
	}
	/*
	 * Reset to a known state if we know what type it is
	 */
	if (info->flash_id != FLASH_UNKNOWN) {
		addr = (vu_short *)info->start[0];
	/* reset bank to read array mode */
		*addr = 0x00FF;
	}

	return (info->size);
}


/*-----------------------------------------------------------------------
 */

int	flash_erase (flash_info_t *info, int s_first, int s_last)
{
	int flag, prot, sect;
	ulong start, now, last;

	if ((s_first < 0) || (s_first > s_last)) {
		if (info->flash_id == FLASH_UNKNOWN) {
			printf ("- missing\n");
		} else {
			printf ("- no sectors to erase\n");
		}
		return 1;
	}

	if (((info->flash_id & FLASH_VENDMASK) != FLASH_MAN_STM) &&
	    ((info->flash_id & FLASH_VENDMASK) != FLASH_MAN_INTEL)) {
		printf ("Can't erase unknown flash type %08lx - aborted\n",
			info->flash_id);
		return 1;
	}

	prot = 0;
	for (sect=s_first; sect<=s_last; ++sect) {
		if (info->protect[sect]) {
			prot++;
		}
	}

	if (prot) {
		printf ("- Warning: %d protected sectors will not be erased!\n",
			prot);
	} else {
		printf ("\n");
	}

	start = get_timer (0);
	last  = start;

	/* Start erase on unprotected sectors */
	for (sect = s_first; sect<=s_last; sect++) {
		if (info->protect[sect] == 0) {	/* not protected */
			vu_short *addr = (vu_short *)(info->start[sect]);
			/* Disable interrupts which might cause a timeout here */
			flag = disable_interrupts();

			/* Single Block Erase Command */
			*addr = 0x0020;
			/* Confirm */
			*addr = 0x00D0;
			/* Resume Command, as per errata update */
			*addr = 0x00D0;

			/* re-enable interrupts if necessary */
			if (flag)
				enable_interrupts();

			/* wait at least 80us - let's wait 1 ms */
			udelay (1000);

			while ((*addr & 0x0080) != 0x0080) {
				if ((now=get_timer(start)) > CFG_FLASH_ERASE_TOUT) {
					printf ("Timeout\n");
					*addr = 0x00FF;		/* reset bank */
					return 1;
				}
				/* show that we're waiting */
				if ((now - last) > 1000) {	/* every second */
					putc ('.');
					last = now;
				}
			}

			/* reset to read mode */
			*addr = 0x00FF;
		}
	}

	printf (" done\n");
	return 0;
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
	ulong cp, wp, data;
	int i, l, rc;

	wp = (addr & ~3);	/* get lower word aligned address */

	/*
	 * handle unaligned start bytes
	 */
	if ((l = addr - wp) != 0) {
		data = 0;
		for (i=0, cp=wp; i<l; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}
		for (; i<4 && cnt>0; ++i) {
			data = (data << 8) | *src++;
			--cnt;
			++cp;
		}
		for (; cnt==0 && i<4; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}

		if ((rc = write_word(info, wp, data)) != 0) {
			return (rc);
		}
		wp += 4;
	}

	/*
	 * handle word aligned part
	 */
	while (cnt >= 4) {
		data = 0;
		for (i=0; i<4; ++i) {
			data = (data << 8) | *src++;
		}
		if ((rc = write_word(info, wp, data)) != 0) {
			return (rc);
		}
		wp  += 4;
		cnt -= 4;
	}

	if (cnt == 0) {
		return (0);
	}

	/*
	 * handle unaligned tail bytes
	 */
	data = 0;
	for (i=0, cp=wp; i<4 && cnt>0; ++i, ++cp) {
		data = (data << 8) | *src++;
		--cnt;
	}
	for (; i<4; ++i, ++cp) {
		data = (data << 8) | (*(uchar *)cp);
	}

	return (write_word(info, wp, data));
}

/*-----------------------------------------------------------------------
 * Write a word to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 * 3 - Block locked or programming failure
 */

static int write_word (flash_info_t *info, ulong dest, ulong data)
{
	int result;

	if ((result = write_halfword (info, dest, (data & ~FLASH_DATA_MASK) >> 16)) == 0)
	{
		dest += 2;
		data = data & FLASH_DATA_MASK;
		result = write_halfword (info, dest, data);
	}
	return result;
}

/*-----------------------------------------------------------------------
 * Write a half word to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 * 3 - Block locked or programming failure
 */
static int write_halfword (flash_info_t *info, ulong dest, ushort data)
{
	vu_short *addr = (vu_short *)dest;
	ulong start;
	int flag;

	/* Check if Flash is (sufficiently) erased */
	if ((*((vu_short *)dest) & data) != data) {
		return (2);
	}

	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts();

	/* Write Command */
	*addr = 0x0010;
	*((vu_short *)dest) = data;

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	/* data polling for D7 */
	flag = 0;
	start = get_timer (0);
	while ((*((vu_short *)dest) & 0x0080) != 0x0080) {
		if (get_timer(start) > CFG_FLASH_WRITE_TOUT) {
			flag = 1;
			break;
		}
	}
	/* check status bits (program error & locked block) */
	if (*((vu_short *)dest) & 0x0012)
	    flag = 3;

	/* Clear Status Registers Command */
	*addr = 0x0050;
	/* Reset to read array mode */
	*addr = 0x00FF;

	return (flag);

}

#ifdef CFG_FLASH_PROTECTION

/*-----------------------------------------------------------------------
 */
static int _flash_protect(flash_info_t *info, ulong sector)
{
    int i;
    int flag;
    ulong status;
    int rcode = 0;
    vu_short *addr = (vu_short *)sector;

    switch(info->flash_id & FLASH_TYPEMASK) {
        case FLASH_STMW320CB:
        case FLASH_STMW640ECB:
        case FLASH_28F320C3B:
        case FLASH_28F640C3B:
            /* Disable interrupts which might cause Flash to timeout */
            flag = disable_interrupts();

            /* Issue command */
            *addr = 0x0050;             /* Clear the status register */
            *addr = 0x0060;             /* Set lock bit setup */
            *addr = 0x0001;             /* Set lock bit confirm */
            *addr = 0x0090;             /* check new state set correctly */
            status = *(addr + 2);
            if ((status & 0x0001) != 0x0001) {
                printf("Protect %x sector failed: %x\n",
                       (uint)sector, (uint)status);
                rcode = 1;
            }
    
            /* Restore read mode */
            *addr = 0x00ff;

            /* re-enable interrupts if necessary */
            if (flag)
                enable_interrupts();

            break;
    }

    if (!rcode) {
    /* Turn protection on for this sector */
        for (i = 0; i < info->sector_count; i++) {
            if (info->start[i] == sector) {
                info->protect[i] = 1;
                break;
            }
        }
    }

    return rcode;
} /* end _flash_protect() */

static int _flash_unprotect(flash_info_t *info, ulong sector)
{
    int i;
    int flag;
    ulong status;
    int rcode = 0;
    vu_short *addr = (vu_short *)sector;

    switch(info->flash_id & FLASH_TYPEMASK) {
        case FLASH_STMW320CB:
        case FLASH_STMW640ECB:
        case FLASH_28F320C3B:
        case FLASH_28F640C3B:
            /* Disable interrupts which might cause Flash to timeout */
            flag = disable_interrupts();

            *addr = 0x0050;             /* Clear the status register */
            *addr = 0x0060;             /* Clear lock bit setup */
            *addr = 0x00D0;             /* Clear lock bit confirm */
            *addr = 0x0090;             /* check new state set correctly */
            status = *(addr + 2);
            if ((status & 0x0001) != 0x0000) {
                printf("Un-protect %x sector failed: %x\n",
                       (uint)sector, (uint)status);
                rcode = 1;
            }
    
            /* restore read mode */
            *addr = 0x00ff;

            /* re-enable interrupts if necessary */
            if (flag)
                enable_interrupts();

            break;
    }


    if (!rcode) {
        for (i = 0; i < info->sector_count; i++) {
            if (info->start[i] == sector) {
                info->protect[i] = 0;
                break;
            }
        }
    }


    return rcode;
} /* end _flash_unprotect() */


int flash_real_protect(flash_info_t *info, long sector, int prot)
{
    int rcode;

    if (prot)
        rcode = _flash_protect(info, info->start[sector]);
    else
        rcode = _flash_unprotect(info, info->start[sector]);

    return rcode;
} /* end flash_real_protect() */

/*-----------------------------------------------------------------------
 */

#endif /* CFG_FLASH_PROTECTION */

