/*
*
* Do not edit $tags$ manually, CVS will take care of this!
*
* Copyright (c) 2003 Tait Electronics Ltd, All rights reserved.
*
* Source File Name:   $RCSfile: logdump.c,v $
*
* Module Name:        Dump kernel log from RAM
*
* Sub-System Name:    U-boot external program
*
* Description:
* Quicky prog to find and print null terminated string from RAM at the address we expect
* the static kernel log to be after a reset
* Revision:            $Revision: 1.1 $
*
* Date:                $Date: 2004/12/08 03:03:42 $
*
* Author:              Rob Searle, Seung Lee, Dmytro Bablinyuk, Robert Ancell, Robin Gilks
*
* Locked By:           $Locker:  $
*
* This file has never been reviewed
*
*/

#include <asm/byteorder.h>

#include <common.h>
#include <exports.h>
#include <logbuff.h>



/*
* NOTE
* Order of code is important here - the bootloader external programs
* expect the entry point to be at the start of the program and to
* have the correct name.
*/

unsigned long my_strtoul(const char *cp);


int logdump(int argc, char *argv[]){

    long address = 0;
    int i;
    char *s, c;
      
    app_startup(argv);

    if (argc != 2){
	printf("Usage: logdump <address>\n", argv[0]);
        return 1;
    }

  
    address = my_strtoul(argv[1]);
    for (i=0; i < LOGBUFF_MASK; i++) {
	s = address + i;
	c = *s;
	if (c == '\0')
	    break;
	c &= 0x7f;
	if ((c >= 0x20) || (c = '\n'))
	    putc (c);
    }
    return 0;


}


/* very simple ctype for isdigit etc */

#define _U	0x01	/* upper */
#define _L	0x02	/* lower */
#define _D	0x04	/* digit */
#define _X	0x08	/* hex digit */


#define __ismask(x) (_ctype[(int)(unsigned char)(x)])

#define isdigit(c)	((__ismask(c)&(_D)) != 0)
#define islower(c)	((__ismask(c)&(_L)) != 0)
#define isxdigit(c)	((__ismask(c)&(_D|_X)) != 0)


unsigned char _ctype[] = {
0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,
_D,_D,_D,_D,_D,_D,_D,_D,			/* 48-55 */
_D,_D,0,0,0,0,0,0,
0,_U|_X,_U|_X,_U|_X,_U|_X,_U|_X,_U|_X,_U,	/* 64-71 */
0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,
0,_L|_X,_L|_X,_L|_X,_L|_X,_L|_X,_L|_X,_L,	/* 96-103 */
0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0};


static inline unsigned char toupper(unsigned char c)
{
	if (islower(c))
		c -= 'a'-'A';
	return c;
}




unsigned long my_strtoul(const char *cp)
{
    unsigned long result = 0,value;
    int base = 16;

    while (isxdigit(*cp) && (value = isdigit(*cp) ? *cp-'0' : (islower(*cp)
	    ? toupper(*cp) : *cp)-'A'+10) < base) {
	result = result*base + value;
	cp++;
    }
    return result;
}

